# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

import typing

def find_subgraphs(
    inputs: typing.Sequence[typing.Sequence[str]],
    output: typing.Sequence[str],
    size_dict: typing.Mapping[str, float],
) -> list[list[int]]:
    r"""
    Find all disconnected subgraphs of a specified contraction.
    """
    ...

def optimize_greedy(
    inputs: typing.Sequence[typing.Sequence[str]],
    output: typing.Sequence[str],
    size_dict: typing.Mapping[str, float],
    costmod: typing.Optional[float] = None,
    temperature: typing.Optional[float] = None,
    seed: typing.Optional[int] = None,
    simplify: typing.Optional[bool] = None,
    use_ssa: typing.Optional[bool] = None,
) -> list[list[int]]:
    r"""
    Find a contraction path using a (randomizable) greedy algorithm.

    Parameters
    ----------
    inputs : Sequence[Sequence[str]]
        The indices of each input tensor.
    output : Sequence[str]
        The indices of the output tensor.
    size_dict : dict[str, int]
        A dictionary mapping indices to their dimension.
    costmod : float, optional
        When assessing local greedy scores how much to weight the size of the
        tensors removed compared to the size of the tensor added::

            score = size_ab / costmod - (size_a + size_b) * costmod

        This can be a useful hyper-parameter to tune.
    temperature : float, optional
        When asessing local greedy scores, how much to randomly perturb the
        score. This is implemented as::

            score -> sign(score) * log(|score|) - temperature * gumbel()

        which implements boltzmann sampling.
    simplify : bool, optional
        Whether to perform simplifications before optimizing. These are:

        - ignore any indices that appear in all terms
        - combine any repeated indices within a single term
        - reduce any non-output indices that only appear on a single term
        - combine any scalar terms
        - combine any tensors with matching indices (hadamard products)

        Such simpifications may be required in the general case for the proper
        functioning of the core optimization, but may be skipped if the input
        indices are already in a simplified form.
    use_ssa : bool, optional
        Whether to return the contraction path in 'single static assignment'
        (SSA) format (i.e. as if each intermediate is appended to the list of
        inputs, without removals). This can be quicker and easier to work with
        than the 'linear recycled' format that `numpy` and `opt_einsum` use.

    Returns
    -------
    path : list[list[int]]
        The contraction path, given as a sequence of pairs of node indices. It
        may also have single term contractions if `simplify=True`.
    """
    ...

def optimize_optimal(
    inputs: typing.Sequence[typing.Sequence[str]],
    output: typing.Sequence[str],
    size_dict: typing.Mapping[str, float],
    minimize: typing.Optional[str] = None,
    cost_cap: typing.Optional[float] = None,
    search_outer: typing.Optional[bool] = None,
    simplify: typing.Optional[bool] = None,
    use_ssa: typing.Optional[bool] = None,
) -> list[list[int]]:
    r"""
    Find an optimal contraction ordering.

    Parameters
    ----------
    inputs : Sequence[Sequence[str]]
        The indices of each input tensor.
    output : Sequence[str]
        The indices of the output tensor.
    size_dict : dict[str, int]
        The size of each index.
    minimize : str, optional
        The cost function to minimize. The options are:

        - "flops": minimize with respect to total operation count only
          (also known as contraction cost)
        - "size": minimize with respect to maximum intermediate size only
          (also known as contraction width)
        - 'max': minimize the single most expensive contraction, i.e. the
          asymptotic (in index size) scaling of the contraction
        - 'write' : minimize the sum of all tensor sizes, i.e. memory written
        - 'combo' or 'combo={factor}` : minimize the sum of
          FLOPS + factor * WRITE, with a default factor of 64.
        - 'limit' or 'limit={factor}` : minimize the sum of
          MAX(FLOPS, alpha * WRITE) for each individual contraction, with a
          default factor of 64.

        'combo' is generally a good default in term of practical hardware
        performance, where both memory bandwidth and compute are limited.
    cost_cap : float, optional
        The maximum cost of a contraction to initially consider. This acts like
        a sieve and is doubled at each iteration until the optimal path can
        be found, but supplying an accurate guess can speed up the algorithm.
    search_outer : bool, optional
        If True, consider outer product contractions. This is much slower but
        theoretically might be required to find the true optimal 'flops'
        ordering. In practical settings (i.e. with minimize='combo'), outer
        products should not be required.
    simplify : bool, optional
        Whether to perform simplifications before optimizing. These are:

        - ignore any indices that appear in all terms
        - combine any repeated indices within a single term
        - reduce any non-output indices that only appear on a single term
        - combine any scalar terms
        - combine any tensors with matching indices (hadamard products)

        Such simpifications may be required in the general case for the proper
        functioning of the core optimization, but may be skipped if the input
        indices are already in a simplified form.
    use_ssa : bool, optional
        Whether to return the contraction path in 'single static assignment'
        (SSA) format (i.e. as if each intermediate is appended to the list of
        inputs, without removals). This can be quicker and easier to work with
        than the 'linear recycled' format that `numpy` and `opt_einsum` use.

    Returns
    -------
    path : list[list[int]]
         The contraction path, given as a sequence of pairs of node indices. It
         may also have single term contractions if `simplify=True`.
    """
    ...

def optimize_random_greedy_track_flops(
    inputs: typing.Sequence[typing.Sequence[str]],
    output: typing.Sequence[str],
    size_dict: typing.Mapping[str, float],
    ntrials: int,
    costmod: typing.Optional[tuple[float, float]] = None,
    temperature: typing.Optional[tuple[float, float]] = None,
    seed: typing.Optional[int] = None,
    simplify: typing.Optional[bool] = None,
    use_ssa: typing.Optional[bool] = None,
) -> tuple[list[list[int]], float]:
    r"""
    Perform a batch of random greedy optimizations, simulteneously tracking
    the best contraction path in terms of flops, so as to avoid constructing a
    separate contraction tree.

    Parameters
    ----------
    inputs : tuple[tuple[str]]
        The indices of each input tensor.
    output : tuple[str]
        The indices of the output tensor.
    size_dict : dict[str, int]
        A dictionary mapping indices to their dimension.
    ntrials : int, optional
        The number of random greedy trials to perform. The default is 1.
    costmod : (float, float), optional
        When assessing local greedy scores how much to weight the size of the
        tensors removed compared to the size of the tensor added::

            score = size_ab / costmod - (size_a + size_b) * costmod

        It is sampled uniformly from the given range.
    temperature : (float, float), optional
        When asessing local greedy scores, how much to randomly perturb the
        score. This is implemented as::

            score -> sign(score) * log(|score|) - temperature * gumbel()

        which implements boltzmann sampling. It is sampled log-uniformly from
        the given range.
    seed : int, optional
        The seed for the random number generator.
    simplify : bool, optional
        Whether to perform simplifications before optimizing. These are:

        - ignore any indices that appear in all terms
        - combine any repeated indices within a single term
        - reduce any non-output indices that only appear on a single term
        - combine any scalar terms
        - combine any tensors with matching indices (hadamard products)

        Such simpifications may be required in the general case for the proper
        functioning of the core optimization, but may be skipped if the input
        indices are already in a simplified form.
    use_ssa : bool, optional
        Whether to return the contraction path in 'single static assignment'
        (SSA) format (i.e. as if each intermediate is appended to the list of
        inputs, without removals). This can be quicker and easier to work with
        than the 'linear recycled' format that `numpy` and `opt_einsum` use.

    Returns
    -------
    path : list[list[int]]
        The best contraction path, given as a sequence of pairs of node
        indices.
    flops : float
        The flops (/ contraction cost / number of multiplications), of the best
        contraction path, given log10.
    """
    ...

def optimize_simplify(
    inputs: typing.Sequence[typing.Sequence[str]],
    output: typing.Sequence[str],
    size_dict: typing.Mapping[str, float],
    use_ssa: typing.Optional[bool] = None,
) -> list[list[int]]:
    r"""
    Find the (partial) contracton path for simplifiactions only.

    Parameters
    ----------
    inputs : Sequence[Sequence[str]]
        The indices of each input tensor.
    output : Sequence[str]
        The indices of the output tensor.
    size_dict : dict[str, int]
        A dictionary mapping indices to their dimension.
    use_ssa : bool, optional
        Whether to return the contraction path in 'single static assignment'
        (SSA) format (i.e. as if each intermediate is appended to the list of
        inputs, without removals). This can be quicker and easier to work with
        than the 'linear recycled' format that `numpy` and `opt_einsum` use.

    Returns
    -------
    path : list[list[int]]
        The contraction path, given as a sequence of pairs of node indices. It
        may also have single term contractions.
    """
    ...

def ssa_to_linear(
    ssa_path: typing.Sequence[typing.Sequence[int]], n: typing.Optional[int] = None
) -> list[list[int]]:
    r"""
    Convert a SSA path to linear format.
    """
    ...



import pytest

try:
    import cotengra as ctg

    ctg_missing = False
except ImportError:
    ctg_missing = True
    ctg = None

import cotengrust as ctgr


requires_cotengra = pytest.mark.skipif(ctg_missing, reason="requires cotengra")


@pytest.mark.parametrize("which", ["greedy", "optimal"])
def test_basic_call(which):
    inputs = [('a', 'b'), ('b', 'c'), ('c', 'd'), ('d', 'a')]
    output = ('b', 'd')
    size_dict = {'a': 2, 'b': 3, 'c': 4, 'd': 5}
    path = {
        "greedy": ctgr.optimize_greedy,
        "optimal": ctgr.optimize_optimal,
    }[
        which
    ](inputs, output, size_dict)
    assert all(len(con) <= 2 for con in path)


def find_output_str(lhs):
    tmp_lhs = lhs.replace(",", "")
    return "".join(s for s in sorted(set(tmp_lhs)) if tmp_lhs.count(s) == 1)


def eq_to_inputs_output(eq):
    if "->" not in eq:
        eq += "->" + find_output_str(eq)
    inputs, output = eq.split("->")
    inputs = inputs.split(",")
    inputs = [list(s) for s in inputs]
    output = list(output)
    return inputs, output


def get_rand_size_dict(inputs, d_min=2, d_max=3):
    import random

    size_dict = {}
    for term in inputs:
        for ix in term:
            if ix not in size_dict:
                size_dict[ix] = random.randint(d_min, d_max)
    return size_dict


# these are taken from opt_einsum
test_case_eqs = [
    # Test single-term equations
    "->",
    "a->a",
    "ab->ab",
    "ab->ba",
    "abc->bca",
    "abc->b",
    "baa->ba",
    "aba->b",
    # Test scalar-like operations
    "a,->a",
    "ab,->ab",
    ",ab,->ab",
    ",,->",
    # Test hadamard-like products
    "a,ab,abc->abc",
    "a,b,ab->ab",
    # Test index-transformations
    "ea,fb,gc,hd,abcd->efgh",
    "ea,fb,abcd,gc,hd->efgh",
    "abcd,ea,fb,gc,hd->efgh",
    # Test complex contractions
    "acdf,jbje,gihb,hfac,gfac,gifabc,hfac",
    "cd,bdhe,aidb,hgca,gc,hgibcd,hgac",
    "abhe,hidj,jgba,hiab,gab",
    "bde,cdh,agdb,hica,ibd,hgicd,hiac",
    "chd,bde,agbc,hiad,hgc,hgi,hiad",
    "chd,bde,agbc,hiad,bdi,cgh,agdb",
    "bdhe,acad,hiab,agac,hibd",
    # Test collapse
    "ab,ab,c->",
    "ab,ab,c->c",
    "ab,ab,cd,cd->",
    "ab,ab,cd,cd->ac",
    "ab,ab,cd,cd->cd",
    "ab,ab,cd,cd,ef,ef->",
    # Test outer prodcuts
    "ab,cd,ef->abcdef",
    "ab,cd,ef->acdf",
    "ab,cd,de->abcde",
    "ab,cd,de->be",
    "ab,bcd,cd->abcd",
    "ab,bcd,cd->abd",
    # Random test cases that have previously failed
    "eb,cb,fb->cef",
    "dd,fb,be,cdb->cef",
    "bca,cdb,dbf,afc->",
    "dcc,fce,ea,dbf->ab",
    "fdf,cdd,ccd,afe->ae",
    "abcd,ad",
    "ed,fcd,ff,bcf->be",
    "baa,dcf,af,cde->be",
    "bd,db,eac->ace",
    "fff,fae,bef,def->abd",
    "efc,dbc,acf,fd->abe",
    # Inner products
    "ab,ab",
    "ab,ba",
    "abc,abc",
    "abc,bac",
    "abc,cba",
    # GEMM test cases
    "ab,bc",
    "ab,cb",
    "ba,bc",
    "ba,cb",
    "abcd,cd",
    "abcd,ab",
    "abcd,cdef",
    "abcd,cdef->feba",
    "abcd,efdc",
    # Inner than dot
    "aab,bc->ac",
    "ab,bcc->ac",
    "aab,bcc->ac",
    "baa,bcc->ac",
    "aab,ccb->ac",
    # Randomly built test caes
    "aab,fa,df,ecc->bde",
    "ecb,fef,bad,ed->ac",
    "bcf,bbb,fbf,fc->",
    "bb,ff,be->e",
    "bcb,bb,fc,fff->",
    "fbb,dfd,fc,fc->",
    "afd,ba,cc,dc->bf",
    "adb,bc,fa,cfc->d",
    "bbd,bda,fc,db->acf",
    "dba,ead,cad->bce",
    "aef,fbc,dca->bde",
]


@requires_cotengra
@pytest.mark.parametrize("eq", test_case_eqs)
@pytest.mark.parametrize("which", ["greedy", "optimal"])
def test_manual_cases(eq, which):
    inputs, output = eq_to_inputs_output(eq)
    size_dict = get_rand_size_dict(inputs)
    path = {
        "greedy": ctgr.optimize_greedy,
        "optimal": ctgr.optimize_optimal,
    }[
        which
    ](inputs, output, size_dict)
    assert all(len(con) <= 2 for con in path)
    tree = ctg.ContractionTree.from_path(
        inputs, output, size_dict, path=path, check=True
    )
    assert tree.is_complete()


@requires_cotengra
@pytest.mark.parametrize("seed", range(10))
@pytest.mark.parametrize("which", ["greedy", "optimal"])
def test_basic_rand(seed, which):
    inputs, output, shapes, size_dict = ctg.utils.rand_equation(
        n=10,
        reg=4,
        n_out=2,
        n_hyper_in=1,
        n_hyper_out=1,
        d_min=2,
        d_max=3,
        seed=seed,
    )
    path = {
        "greedy": ctgr.optimize_greedy,
        "optimal": ctgr.optimize_optimal,
    }[
        which
    ](inputs, output, size_dict)
    assert all(len(con) <= 2 for con in path)
    tree = ctg.ContractionTree.from_path(
        inputs, output, size_dict, path=path, check=True
    )
    assert tree.is_complete()


@requires_cotengra
def test_optimal_lattice_eq():
    inputs, output, _, size_dict = ctg.utils.lattice_equation(
        [4, 5], d_max=2, seed=42
    )

    path = ctgr.optimize_optimal(inputs, output, size_dict, minimize='flops')
    tree = ctg.ContractionTree.from_path(
        inputs, output, size_dict, path=path
    )
    assert tree.is_complete()
    assert tree.contraction_cost() == 964

    path = ctgr.optimize_optimal(inputs, output, size_dict, minimize='size')
    assert all(len(con) <= 2 for con in path)
    tree = ctg.ContractionTree.from_path(
        inputs, output, size_dict, path=path
    )
    assert tree.contraction_width() == pytest.approx(5)


@requires_cotengra
def test_optimize_random_greedy_log_flops():
    inputs, output, _, size_dict = ctg.utils.lattice_equation(
        [10, 10], d_max=3, seed=42
    )

    path, cost1 = ctgr.optimize_random_greedy_track_flops(
        inputs, output, size_dict, ntrials=4, seed=42
    )
    _, cost2 = ctgr.optimize_random_greedy_track_flops(
        inputs, output, size_dict, ntrials=4, seed=42
    )
    assert cost1 == cost2
    tree = ctg.ContractionTree.from_path(
        inputs, output, size_dict, path=path
    )
    assert tree.is_complete()
    assert tree.contraction_cost(log=10) == pytest.approx(cost1)